/*   FILE: common.c --
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 18 July 2004
 *
 * Copyright (C) 2004 W. Michael Petullo <mike@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <assert.h>
#include <unistd.h>
#include <stdarg.h>
#include <gnome-keyring.h>

static GMainLoop *loop = NULL;
static gboolean debug = FALSE;

/* ============================ set_debug () =============================== */ 
/* INPUT: TRUE or FALSE
 * SIDE AFFECT: debugging is turned on or off
 */
void set_debug(gboolean state)
{
	debug = state;
}

/* ============================ l0g () ===================================== */
/* INPUT: similar to printf; all args are valid strings != NULL
 * SIDE AFFECTS: format + args are logged and displayed */
void l0g(const char *format, ...)
{
	/* Used to log issues that cause pam_keyring to fail. */
	va_list args;

	assert(format != NULL);

	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
	va_start(args, format);
/* This code needs root priv? */
	vsyslog(LOG_USER | LOG_ERR, format, args);
/* end root priv. */
	va_end(args);
}

/* ============================ w4rn () ==================================== */
/* INPUT: similar to printf; all args are valid string != NULL
 * SIDE AFFECTS: format + args are logged and displayed iff debug == 1
 * NOTE: Used to log informational messages and issues that should not cause
 *       pam_keyring to fail. */
void w4rn(const char *format, ...)
{
	assert(format != NULL);

	if (debug == TRUE) {
		va_list args;
		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
		va_start(args, format);
/* This code needs root priv? */
		vsyslog(LOG_USER | LOG_ERR, format, args);
/* end root priv. */
		va_end(args);
	}
}

/* ============================ strip_cr () ================================ */
/* INPUT: s, a string
 * SIDE AFFECTS: terminating carriage return in s is removed if it exists
 * OUTPUT: s
 */
char *strip_cr(char *s)
{
	int len;
	assert(s != NULL);
	
	len = strlen(s);
	s[len - 1] = s[len - 1] == '\n' ? 0x00 : s[len - 1];
	return s;
}

/* ============================ log_output () ============================== */
/* INPUT: fd, a valid file descriptor
 * SIDE AFFECTS: contents of fd are logged, usually fd is connected by a 
 *               pipe to another process's stdout or stderr
 */
void log_output(int fd)
{
	FILE *fp;
	char buf[BUFSIZ + 1];
	/* FIXME: check fdopen's retval */
	if ((fp = fdopen(fd, "r")) == NULL) {
		w4rn("pam_keyring: error opening file: %s\n",
		     strerror(errno));
		return;
	}
	while (fgets(buf, BUFSIZ + 1, fp))
		w4rn("pam_keyring: %s\n", buf);
}
