/*   FILE: pam_keyring.c
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 4 June 2004
 *
 * Copyright (C) 2004 W. Michael Petullo <mike@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 */

#include <common.h>
#include <config.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <glib.h>
#include <pwd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <gnome-keyring.h>

#define PAM_SM_SESSION
#define PAM_SM_PASSWORD
#include <security/pam_modules.h>

#ifdef HAVE_SELINUX_SELINUX_H
#include <selinux/selinux.h>
#endif

/* ============================ auth_type_t ================================ */
typedef enum auth_type_t {
	GET_PASS,
	USE_FIRST_PASS,
	TRY_FIRST_PASS
} auth_type_t;

/* ============================ pam_args_t ================================= */
typedef struct pam_args_t {
	auth_type_t auth_type;
} pam_args_t;

/* ============================ preexec_t ================================== */
typedef struct preexec_t {
	char *user;
	char *evar;
} preexec_t;

/* ============================ globals ==================================== */
pam_args_t args;

/* ============================ free_cb () ================================= */
/* INPUT: pamh; data, must point to malloced memory of be NULL; errcode
 * SIDE AFFECTS: if data does not point to NULL then it is freed
 * NOTE: this is registered as a PAM callback function */
void free_cb(pam_handle_t * pamh, void *data, int errcode)
{
	if (data) {
		g_free(data);
	}
}

/* ============================ parse_pam_args () ========================== */
/* INPUT: argc and argv, standard main()-type arguments
 * SIDE AFFECTS: gloabl args is initialized, based on argc and argv */
void parse_pam_args(int argc, const char **argv)
{
	int i;

	assert(argc >= 0);
	for (i = 0; i < argc; i++)
		assert(argv[i]);

	/* first, set default values */
	args.auth_type = GET_PASS;
	set_debug(FALSE);
	for (i = 0; i < argc; i++) {
		w4rn("pam_keyring: pam_sm_open_session args: %s\n",
		     argv[i]);
		if (!strcmp("use_first_pass", argv[i]))
			args.auth_type = USE_FIRST_PASS;
		else if (!strcmp("try_first_pass", argv[i]))
			args.auth_type = TRY_FIRST_PASS;
		else if (!strcmp("debug", argv[i]))
			set_debug(TRUE);
		else
			w4rn("pam_keyring: %s\n",
			     "bad pam_keyring option");
	}
}

/* ============================ converse () ================================ */
/* INPUT: pamh; nargs; message, a prompt message
 * SIDE AFFECTS: resp points to PAM's (user's) response to message
 * OUTPUT: any PAM error code encountered or PAM_SUCCESS
 * NOTE:   adapted from pam_unix/support.c */
static int
converse(pam_handle_t * pamh, int nargs,
	 const struct pam_message **message, struct pam_response **resp)
{
	int retval;
	struct pam_conv *conv;

	assert(pamh != NULL);
	assert(nargs >= 0);
	assert(resp != NULL);

	*resp = NULL;
	retval = pam_get_item(pamh, PAM_CONV, (const void **) &conv);
	if (retval == PAM_SUCCESS)
		retval =
		    conv->conv(nargs, message, resp, conv->appdata_ptr);
	if (!*resp)
		retval = PAM_AUTH_ERR;

	assert(retval != PAM_SUCCESS
	       || (resp != NULL && *resp != NULL
		   && (*resp)->resp != NULL));

	return retval;		/* propagate error status */
}

/* ============================ read_password () =========================== */
/* INPUT: pamh; prompt1, a prompt message
 * SIDE AFFECTS: pass points to PAM's (user's) response to prompt1 (malloc'ed)
 * OUTPUT: any PAM error code encountered or PAM_SUCCESS
 * NOTE:   adapted from pam_unix/support.c (_unix_read_password)
 */
int read_password(pam_handle_t * pamh, const char *prompt1, char **pass)
{
	int retval;
	struct pam_message msg;
	const struct pam_message *pmsg = &msg;
	struct pam_response *resp = NULL;

	assert(pamh != NULL);
	assert(prompt1 != NULL);
	assert(pass != NULL);

	*pass = NULL;
	msg.msg_style = PAM_PROMPT_ECHO_OFF;
	msg.msg = prompt1;
	retval = converse(pamh, 1, &pmsg, &resp);
	if (retval == PAM_SUCCESS)
		*pass = strdup(resp->resp);

	assert(retval != PAM_SUCCESS || (pass != NULL && *pass != NULL));

	return retval;
}

/* ============================ pipewrite () =============================== */
/* INPUT: fd, a valid file descriptor; buf, a buffer of size count
 * SIDE AFFECTS: buf is written to fd
 * OUTPUT: number of bytes written or 0 on error
 * NOTE: SIGPIPE is ignored during this operation to avoid "broken pipe"
 */
int pipewrite(int fd, const void *buf, size_t count)
{
	int fnval;
	struct sigaction ignoresact = {
		.sa_handler = SIG_IGN
	}, oldsact;

	assert(fd >= 0);
	assert(buf != NULL);
	assert(count >= 0);

	/* avoid bomb on command exiting before data written */
	if (sigaction(SIGPIPE, &ignoresact, &oldsact) < 0) {
		fnval = -1;
		goto _return;
	}
	fnval = write(fd, buf, count);
	/* restore old handler */
	if (sigaction(SIGPIPE, &oldsact, NULL) < 0) {
		fnval = -1;
		goto _return;
	}
      _return:
	return fnval;
}

/* ============================ preexec () ================================= */
/* INPUT: user, a valid user name
 * SIDE AFFECTS: the calling process's uid, gid and euid are set to user's
 *               the calling process's GNOME_KEYRING_SOCKET may be set
 * NOTE: this is required because gnome-keyring-daemon restricts connections
 *       to its owner.
 */
static void preexec(const gpointer data)
{
	char *socket;
#ifdef HAVE_SELINUX_SELINUX_H
	int selinux_enabled = 0;
#endif
	struct passwd *passwd_ent;

	assert(data);
	assert(((preexec_t *)data)->user);

	passwd_ent = getpwnam((char *) ((preexec_t *)data)->user);
	if (!passwd_ent) {
		l0g("pam_keyring: error looking up user information for %s\n", (char *) ((preexec_t *)data)->user);
		exit(EXIT_FAILURE);
	}
	/* must exit because g_spawn_sync does not check return value before
	 * executing argument vector */
	if (setgid(passwd_ent->pw_gid) == -1) {
		l0g("pam_keyring: error setting gid (%s)\n",
		    strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (setuid(passwd_ent->pw_uid) == -1) {
		l0g("pam_keyring: error setting uid (%s)\n",
		    strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (seteuid(passwd_ent->pw_uid) == -1) {
		l0g("pam_keyring: error setting euid: %s\n",
		    strerror(errno));
		exit(EXIT_FAILURE);
	}
#ifdef HAVE_SELINUX_SELINUX_H
	w4rn("pam_keyring: %s\n", "going to perform SELinux operations");
	if ((selinux_enabled = is_selinux_enabled()) == -1) {
		l0g("pam_keyring: error checking for SELinux\n");
		exit(EXIT_FAILURE);
	}
	if (selinux_enabled) {
		char *context = "user_u:user_r:user_t";
		if (setexeccon(context) == -1) {
			l0g("pam_keyring: error setting SELinux context to %s\n", context);
			exit(EXIT_FAILURE);
		}
	}
#endif
	if (((preexec_t *)data)->evar)
		putenv(((preexec_t *)data)->evar);
}

/* ============================ keyring_daemon_stop () ===================== */
/* INPUT: data, a valid preexec_t name; pid, the pid of gnome-keyring-daemon
 * SIDE AFFECTS: a SIGTERM is sent to the gnome-keyring-daemon process
 * OUTPUT: TRUE if the SIGTERM is sent, otherwise FALSE
 */
static int keyring_daemon_stop(const preexec_t *data, pid_t pid)
{
	char *cmd_line;
	gchar **argv;
	GError *err = NULL;
	int ret = TRUE, status;

	assert(data != NULL);
	assert(data->user != NULL);

	cmd_line = g_strdup_printf("%s %d", KILL, pid);

	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		l0g("pam_keyring: error parsing %s\n", cmd_line);
		goto _return;
	}
	if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, preexec,
			  (void *) data, NULL, NULL, &status, &err)) {
		l0g("pam_keyring: failed to run kill: %s\n", err->message);
		g_error_free(err);
		goto _return;
	}
	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
		l0g("pam_keyring: kill failed to execute, exit code: %d\n",
		    WEXITSTATUS(status));
		ret = FALSE;
		goto _return;
	}
      _return:
	return ret;
}

/* ============================ keyring_daemon_start () ==================== */
/* INPUT: pamh, a valid pam_handle_t *; data, a valid user preexec_t
 * SIDE AFFECTS: gnome-keyring-daemon is started
 * OUTPUT: gnome-keyring-daemon's PID or 0 on error
 *         data->evar is set to g-k-d's i/o socket
 */
static pid_t keyring_daemon_start(pam_handle_t * pamh, preexec_t *data)
{
	GError *err = NULL;
	char *standard_out = NULL;
	char **lines;
	gint status = 0;
	long pid;
	pid_t fnval = 0;
	char *pid_str, *end;
	gchar **argv = NULL;

	assert(pamh != NULL);
	assert(data != NULL);
	assert(data->user != NULL);

	if (!g_shell_parse_argv(GNOME_KEYRING_DAEMON, NULL, &argv, &err)) {
		l0g("pam_keyring: error parsing %s\n",
		    GNOME_KEYRING_DAEMON);
		goto _return;
	}
	if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, preexec,
			  (void *) data, &standard_out, NULL, &status,
			  &err)) {
		l0g("pam_keyring: failed to run gnome-keyring-daemon: %s\n", err->message);
		g_error_free(err);
		goto _return;
	}
	if (/* FIXME: WIFEXITED(status) &&
	    WEXITSTATUS(status) == 0 && */ standard_out != NULL) {
		lines = g_strsplit(standard_out, "\n", 3);

		if (lines[0] != NULL &&
		    lines[1] != NULL &&
		    g_str_has_prefix(lines[1], "GNOME_KEYRING_PID=")) {
			pid_str = lines[1] + strlen("GNOME_KEYRING_PID=");
			pid = strtol(pid_str, &end, 10);
			if (end != pid_str) {
				fnval = pid;
				/* set variable for ulock operation */
				data->evar = g_strdup(lines[0]);
				/* FIXME: memory leak */
				/* ensure variable will be set in login session */
				if (pam_putenv
				    (pamh,
				     g_strdup(lines[0])) != PAM_SUCCESS) {
					fnval = 0;
					l0g("pam_keyring: error setting %s\n", lines[0]);
					goto _return;
				}
			}
		}

		g_strfreev(lines);
	} else {
		/* daemon failed for some reason */
		l0g("pam_keyring: gnome-keyring-daemon failed to start correctly, exit code: %d\n", WEXITSTATUS(status));
	}
	g_free(standard_out);
      _return:
	g_strfreev(argv);
	return fnval;
}

/* ============================ get_default () ============================= */ 
char *get_default(const preexec_t *data)
{
	int status;
	GError *err = NULL;
	gchar **argv = NULL;
	char *cmd_line = NULL, *name = NULL;

	assert(data != NULL);
	assert(data->user != NULL);

	cmd_line = g_strconcat(GNOME_KEYRINGTOOL, " getdefault", NULL);
	w4rn("pam_keyring: going to execute %s\n", cmd_line);
	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		l0g("pam_keyring: error parsing %s\n", cmd_line);
		goto _return;
	}
	if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, preexec,
			  (void *) data, &name, NULL, &status, &err)) {
		l0g("pam_keyring: failed to run gnome-keyringtool: %s\n", err->message);
		g_error_free(err);
		goto _return;
	}
	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
		l0g("pam_keyring: gnome-keyring failed to execute, exit code: %d\n", WEXITSTATUS(status));
		name = NULL;
	} else {
		strip_cr(name);
	}
	g_strfreev(argv);
	g_free(cmd_line);

_return:
	return name;
}

/* ============================ unlock () ================================== */
/* INPUT: data and authtok
 * SIDE AFFECTS: user's GNOME keyring is unlocked
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
static int unlock(pam_handle_t * pamh, const preexec_t *data, char *authtok)
{
	gchar **argv = NULL;
	int child_exit;
	pid_t pid = -1;
	GError *err = NULL;
	int ret = PAM_SUCCESS;
	int cstdin = -1, cstderr = -1;
	char *cmd_line = NULL, *name = NULL, *op;

	assert(data);
	assert(data->user);
	assert(authtok);

	if ((name = get_default(data)) == NULL) {
		w4rn("pam_keyring: default keyring does not exist; creating one\n");
		name = "system";
		op = "createdefault";
	} else {
		w4rn("pam_keyring: found default keyring: %s\n", name);
		op = "unlock";
	}
	cmd_line = g_strconcat(GNOME_KEYRINGTOOL, " ", op, " ", name, NULL);
	w4rn("pam_keyring: going to execute %s\n", cmd_line);
	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		l0g("pam_keyring: error parsing %s\n", cmd_line);
		goto _return;
	}
	if (g_spawn_async_with_pipes
	    (NULL, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD,
	     preexec, (void *) data, &pid, &cstdin, NULL, &cstderr,
	     &err) == FALSE) {
		l0g("pam_keyring: error executing %s\n", cmd_line);
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	if (pipewrite(cstdin, authtok, strlen(authtok)) != strlen(authtok)) {
		w4rn("pam_keyring: error writing authtok to gnome-keyring\n");
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	close(cstdin);
	log_output(cstderr);
	if (waitpid(pid, &child_exit, 0) == -1) {
		l0g("pam_keyring: error waiting for child\n");
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	close(cstderr);
	ret = WEXITSTATUS(child_exit) ? PAM_SERVICE_ERR : PAM_SUCCESS;
      _return:
	memset(authtok, 0x00, strlen(authtok));
	g_free(authtok);
	g_strfreev(argv);
	g_free(cmd_line);
	return ret;
}

/* ============================ pam_sm_authenticate () ===================== */
/* INPUT: this function is called by PAM
 * SIDE AFFECTS: user's GNOME keyring is unlocked
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
PAM_EXTERN int
pam_sm_authenticate(pam_handle_t * pamh, int flags,
		    int argc, const char **argv)
{
	pid_t *daemon_pid;
	int ret = PAM_SUCCESS;
	char *authtok = NULL;
	preexec_t data = { NULL, NULL };

	assert(pamh);

	parse_pam_args(argc, argv);
	w4rn("pam_keyring: received order to authenticate\n");

	/* needed because gdm does not prompt for username as login does: */
	if ((ret = pam_get_user(pamh, &data.user, NULL)) != PAM_SUCCESS) {
		l0g("pam_keyring: %s\n", "could not get user");
		/* do NOT return PAM_SERVICE_ERR or root will not be able 
		 * to su to other users */
		goto _return;
	}

	if (! strcmp(data.user, "root")) {
		w4rn("pam_keyring: do nothing for root\n");
		ret = PAM_SUCCESS;
		goto _return;
	}

	if ((data.evar = getenv("GNOME_KEYRING_SOCKET")) != NULL) {
		w4rn("pam_keyring: daemon already exists (%s)\n", data.evar);
		/* FIXME: g_strconcat = memory leak? */
		/* ensure variable will be passed to next session */
		if (pam_putenv (pamh, g_strconcat("GNOME_KEYRING_SOCKET=", 
		                                  data.evar, NULL)) != PAM_SUCCESS) {
			l0g("pam_keyring: error setting GNOME_KEYRING_SOCKET\n");
			ret = PAM_SERVICE_ERR;
			goto _return;
		}
		ret = PAM_SUCCESS;
		goto _return;
	}

	if (args.auth_type != GET_PASS) {	/* get password from PAM system */
		char *ptr = NULL;
		if ((ret =
		     pam_get_item(pamh, PAM_AUTHTOK,
				  (const void **) &ptr)) !=
		    PAM_SUCCESS || !ptr) {
			l0g("pam_keyring: %s\n",
			    "could not get password from PAM system");
			if (args.auth_type == USE_FIRST_PASS)
				goto _return;
		} else
			authtok = g_strdup(ptr);
	}
	if (!authtok) {		/* get password directly */
		if ((ret = read_password(pamh, "password:",
					 &authtok)) != PAM_SUCCESS) {
			l0g("pam_keyring: %s\n",
			    "error trying to read password");
			goto _return;
		}
		/* p_s_i copies to PAM-internal memory */
		if ((ret =
		     pam_set_item(pamh, PAM_AUTHTOK,
				  authtok)) != PAM_SUCCESS) {
			l0g("pam_keyring: %s\n",
			    "error trying to export password");
			goto _return;
		}
	}
	w4rn("pam_keyring: starting gnome-keyring-daemon\n");
	daemon_pid = g_new0(pid_t, 1); // FIXME: memory leak?
	if ((*daemon_pid = keyring_daemon_start(pamh, &data)) == 0) {
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	if ((ret =
	     pam_set_data(pamh, "pam_keyring_gkd_pid", daemon_pid,
			  free_cb)) != PAM_SUCCESS) {
		l0g("pam_keyring: %s\n",
		    "error trying to save gnome-keyring-daemon PID");
		goto _return;
	}
	w4rn("pam_keyring: unlocking keyring\n");
	ret = unlock(pamh, &data, authtok);
      _return:
	return ret;
}

/* ============================ pam_sm_open_session () ===================== */
/* NOTE: placeholder function so PAM does not get mad */
PAM_EXTERN int
pam_sm_open_session(pam_handle_t * pamh, int flags,
		    int argc, const char **argv)
{
	return PAM_SUCCESS;
}

/* ============================ update_keyring_password () ================= */
/* INPUT: pamh, a valid pam_handle_t *; authtok, an authentication token
 * SIDE AFFECTS: the default GNOME keyring's password is set to authtok
 * OUTPUT: PAM_SUCCESS or PAM_SERVICE_ERR
 */
static int update_keyring_password(pam_handle_t * pamh, char *authtok)
{
	int child_exit;
	pid_t pid = -1;
	GError *err = NULL;
	gchar **argv = NULL;
	int ret = PAM_SUCCESS;
	int cstdin = -1, cstderr = -1;
	char *cmd_line = NULL, *name = NULL;
	preexec_t data = { NULL, NULL };
	w4rn("pam_keyring: received order to change password\n");
	/* needed because gdm does not prompt for username as login does: */
	if ((ret = pam_get_user(pamh, &data.user, NULL)) != PAM_SUCCESS) {
		l0g("pam_keyring: %s\n", "could not get user");
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	if ((name = get_default(data.user)) == NULL) {
		w4rn("pam_keyring: default keyring does not exist\n");
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	cmd_line = g_strconcat(GNOME_KEYRINGTOOL, " passwd ", name, NULL);
	w4rn("pam_keyring: going to execute %s\n", cmd_line);
	if (!g_shell_parse_argv(cmd_line, NULL, &argv, &err)) {
		l0g("pam_keyring: error parsing %s\n", cmd_line);
		goto _return;
	}
	if (g_spawn_async_with_pipes
	    (NULL, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD,
	     preexec, (void *) &data, &pid, &cstdin, NULL, &cstderr,
	     &err) == FALSE) {
		l0g("pam_keyring: error executing %s\n", cmd_line);
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	if (pipewrite(cstdin, authtok, strlen(authtok)) != strlen(authtok)) {
		w4rn("pam_keyring: error writing authtok to gnome-keyring\n");
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	close(cstdin);
	log_output(cstderr);
	if (waitpid(pid, &child_exit, 0) == -1) {
		l0g("pam_keyring: error waiting for child\n");
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
	close(cstderr);
	ret = WEXITSTATUS(child_exit) ? PAM_SERVICE_ERR : PAM_SUCCESS;

      _return:
	return ret;
}

/* ============================ pam_sm_chauthtok () ======================== */
/* INPUT: this function is called by PAM, gnome-keyring-daemon must be running
 * SIDE AFFECTS: user's GNOME keyring has its password changed
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t * pamh, int flags, int argc,
		 const char **argv)
{
	int ret = PAM_SUCCESS;

	assert(pamh);

	parse_pam_args(argc, argv);
	if (flags & PAM_PRELIM_CHECK) {
		if (gnome_keyring_is_available())
			ret = PAM_SUCCESS;
		else
			ret = PAM_TRY_AGAIN;
		goto _return;
	} else if (flags & PAM_UPDATE_AUTHTOK) {
		pid_t child;
		int child_exit;
		preexec_t data;
		char *authtok = NULL;
		char *authtok2 = NULL;
		char *oldauthtok = NULL;
		/* FIXME: this code could be shared */
		if (args.auth_type != GET_PASS) {	/* get password from PAM system */
			char *ptr = NULL;
			if ((ret =
			     pam_get_item(pamh, PAM_AUTHTOK,
					  (const void **) &ptr)) !=
			    PAM_SUCCESS || !ptr) {
				l0g("pam_keyring: %s\n",
				    "could not get new password from PAM system");
				if (args.auth_type == USE_FIRST_PASS)
					goto _return;
			} else
				authtok = g_strdup(ptr);
			if ((ret =
			     pam_get_item(pamh, PAM_OLDAUTHTOK,
					  (const void **) &ptr)) !=
			    PAM_SUCCESS || !ptr) {
				l0g("pam_keyring: %s\n",
				    "could not get old password from PAM system");
				if (args.auth_type == USE_FIRST_PASS)
					goto _return;
			} else
				oldauthtok = g_strdup(ptr);
		}
		if (!oldauthtok) {	/* get password directly */
			if ((ret =
			     read_password(pamh,
					   "Current password:",
					   &oldauthtok)) != PAM_SUCCESS) {
				l0g("pam_keyring: %s\n",
				    "error trying to read old password");
				goto _return;
			}
			/* p_s_i copies to PAM-internal memory */
			if ((ret =
			     pam_set_item(pamh, PAM_OLDAUTHTOK,
					  oldauthtok)) != PAM_SUCCESS) {
				l0g("pam_keyring: %s\n",
				    "error trying to export old password");
				goto _return;
			}
		}
		if (!authtok) {	/* get password directly */
			if ((ret =
			     read_password(pamh, "New password:",
					   &authtok)) != PAM_SUCCESS) {
				l0g("pam_keyring: %s\n",
				    "error trying to read new password");
				goto _return;
			}
			if ((ret =
			     read_password(pamh,
					   "Retype new password:",
					   &authtok2)) != PAM_SUCCESS) {
				l0g("pam_keyring: %s\n",
				    "error trying to read password");
				goto _return;
			}
			if (strcmp(authtok, authtok2) != 0) {
				/* FIXME: see make_remark in pam_cracklib */
				l0g("pam_keyring: new password entries did not match\n");
				if ((ret =
				     pam_set_item(pamh,
						  PAM_AUTHTOK,
						  NULL)) != PAM_SUCCESS) {
					l0g("pam_keyring: %s\n",
					    "error trying to blank badly input password");
					goto _return;
				}
				ret = PAM_AUTHTOK_RECOVER_ERR;
				goto _return;
			}
			/* p_s_i copies to PAM-internal memory */
			if ((ret =
			     pam_set_item(pamh, PAM_AUTHTOK,
					  authtok)) != PAM_SUCCESS) {
				l0g("pam_keyring: %s\n",
				    "error trying to export new password");
				goto _return;
			}
		}

		/* FIXME: how to prompt/reprompt for new password ? */

		if ((child = fork()) < 0) {
			l0g("pam_keyring: fork failed\n");
			ret = PAM_SERVICE_ERR;
			goto _return;
		} else if (child == 0) {	/* child */
			if ((ret =
			     pam_get_user(pamh, &data.user,
					  NULL)) != PAM_SUCCESS) {
				l0g("pam_keyring: %s\n",
				    "could not get user");
				/* do NOT return PAM_SERVICE_ERR or root will not be able 
				 * to su to other users */
				goto _return;
			}

			if (! strcmp(data.user, "root")) { /* NOOP for root */
				w4rn("pam_keyring: do nothing for root\n");
				ret = PAM_SUCCESS;
				goto _return;
			}

			preexec(&data);
			/* FIXME: need to split this into a seperate "update" binary. */
			if ((ret =
			     update_keyring_password(pamh,
						     authtok)) !=
			    PAM_SUCCESS) {
				l0g("pam_keyring: error updating keyring password\n");
				exit(EXIT_FAILURE);
			}
			exit(EXIT_SUCCESS);
		} else {
			if (waitpid(child, &child_exit, 0) == -1) {
				l0g("pam_keyring: error waiting for child (%d): %s\n", child, strerror(errno));
				ret = PAM_SERVICE_ERR;
				goto _return;
			}
		}
	}
      _return:
	return ret;
}

/* ============================ pam_sm_close_session () ==================== */
/* INPUT: this function is called by PAM
 * SIDE AFFECTS: user's gnome-keyring-daemon is sent a SIGTERM
 * OUTPUT: PAM error code on error or PAM_SUCCESS
 */
PAM_EXTERN int
pam_sm_close_session(pam_handle_t * pamh, int flags, int argc,
		     const char **argv)
{
	preexec_t data = { NULL, NULL };
	pid_t *daemon_pid;
	int ret = PAM_SUCCESS;

	assert(pamh);

	parse_pam_args(argc, argv);

	w4rn("pam_keyring: received order to close session\n");

	if ((ret = pam_get_user(pamh, &data.user, NULL)) != PAM_SUCCESS) {
		l0g("pam_keyring: %s\n", "could not get user");
		/* do NOT return PAM_SERVICE_ERR or root will not be able 
		 * to su to other users */
		goto _return;
	}
	if (! strcmp(data.user, "root")) {
		w4rn("pam_keyring: do nothing for root\n");
		ret = PAM_SUCCESS;
		goto _return;
	}
	if ((ret =
	     pam_get_data(pamh, "pam_keyring_gkd_pid",
			  (const void **) &daemon_pid)) != PAM_SUCCESS) {
		l0g("pam_keyring: %s\n",
		    "could not retrieve gnome-keyring-deamon PID");
		ret = PAM_SUCCESS;
		goto _return;
	}
	if (keyring_daemon_stop(&data, *daemon_pid) == FALSE) {
		l0g("pam_keyring: error trying to kill gnome-keyring-daemon (%d)\n", *daemon_pid);
		ret = PAM_SERVICE_ERR;
		goto _return;
	}
      _return:
	return ret;
}

/* ============================ pam_sm_setcred () ========================== */
/* NOTE: placeholder function so PAM does not get mad */
PAM_EXTERN int
pam_sm_setcred(pam_handle_t * pamh, int flags, int argc, const char **argv)
{
	return PAM_SUCCESS;
}
